#include"zega.h"
#include<math.h>

QPoint rotate(QPoint point, double deg) {
	QPoint rtnPoint;
	double rad = deg * M_PI / 180.0;
	rtnPoint.setX(point.x() * cos(rad) * -1 + point.y() * sin(rad));
	rtnPoint.setY((point.x() * sin(rad) + point.y() * cos(rad)) * -1);
	return rtnPoint;
}

Config Zega::config = Config("Zega");

Zega::Zega(QWidget* parent, const char* name) :
	QWidget(parent, name) {
	// ȥ
	setCaption(tr("Zega"));

	// ˥塼Сɲ
	_menu = new QMenuBar(this, "menubar");

	_filemenu = new QPopupMenu(this, "filemenu");
	_filemenu->insertItem(tr("&Load File"), this, SLOT(slotLoad()));
	_filemenu->insertSeparator();
	_filemenu->insertItem(tr("&Quit"), _app, SLOT(quit()));

	// Ͽ˥塼
	_scalemenu = new QPopupMenu(this, "scalemenu");
	config.setGroup("Maps");
	if ( !config.hasKey("mapNum") ) {
		config.writeEntry("mapNum", 3);
		config.writeEntry("map0", "150000,maps/1");
		config.writeEntry("map1", "21000,maps/2");
		config.writeEntry("map2", "8000,maps/3");
	}

	int mapNum = config.readNumEntry("mapNum");
	QString strTmp;
	QString strNumTmp;
	int intMapId = 0;
	mapDirs.clear();
	for (int i = 0; i < mapNum; i++) {
		strTmp = QString("map") + strNumTmp.setNum(i);
		if (config.hasKey(strTmp) ) {
			QStringList mapInfo = config.readListEntry(strTmp, ',');
			if (mapInfo.count() >= 2) {
				mapDirs.append(mapInfo[1]);
				int itemId = _scalemenu->insertItem(mapInfo[0], this, SLOT(slotChScale(int)));
				_scalemenu->setItemParameter(itemId, intMapId);
				intMapId++;
			}
		}
	}

	_gpsmenu = new QPopupMenu(this, "gpsmenu");
	_gpsmenu->insertItem(tr("&Start GPS"), this, SLOT(startGps()));
	_gpsmenu->insertItem(tr("S&top GPS"), this, SLOT(stopGps()));

	_helpmenu = new QPopupMenu(this, "helpmenu");
	_helpmenu->insertItem(tr("&About"), this, SLOT(about()));

	_menu->insertItem(tr("&File"), _filemenu);
	_menu->insertItem(tr("&Scale"), _scalemenu);
	_menu->insertItem(tr("&GPS"), _gpsmenu);
	_menu->insertItem(tr("&Help"), _helpmenu);

	config.setGroup("CurrentCoordinate");
	// ֤߰¸ߤʤؤ
	if (!config.hasKey("lat"))
		config.writeEntry("lat", "N35.40.40.00");
	if (!config.hasKey("lon"))
		config.writeEntry("lon", "E139.46.15.00");

	_centerNS.parse(config.readEntry("lat"));
	_centerEW.parse(config.readEntry("lon"));
	_pixelNS = 0;
	_pixelEW = 0;

	// Ͽ
	_mapIndex = 0;
	config.setGroup("Maps");
	if (!config.hasKey("currentMap"))
		config.writeEntry("currentMap", "0");
	if (config.hasKey("currentMap") ) {
		_mapIndex = config.readNumEntry("currentMap");
		if (_mapIndex >= mapDirs.count())
			_mapIndex = 0;
	}
	_mapList = new MapList(mapDirs[_mapIndex]);

	// GPS
	config.setGroup("GPS");
	// GPSեѥ
	if ( !config.hasKey("logDir") )
		config.writeEntry("logDir", "/mnt/card/gpslog");
	// GPSǥХե
	if ( !config.hasKey("devFile") )
		config.writeEntry("devFile", "/dev/ttyS3");

	// ľȤGPS־򵭲ֳ
	if ( !config.hasKey("recentTerm") )
		config.writeEntry("recentTerm", "1");
	int recentTerm = config.readNumEntry("recentTerm");
	if (recentTerm <= 0)
		recentTerm = 1;

	// ľȤGPS־Ͽ
	if ( !config.hasKey("recentLogNum") )
		config.writeEntry("recentLogNum", "600");
	int recentLogNum = config.readNumEntry("recentLogNum");
	if (recentLogNum <= 0)
		recentLogNum = 600;

	// ŤȤGPS־򵭲ֳ
	if ( !config.hasKey("obsolateTerm") )
		config.writeEntry("obsolateTerm", "15");
	int obsolateTerm = config.readNumEntry("obsolateTerm");
	if (obsolateTerm <= recentTerm)
		obsolateTerm = recentTerm * 4;

	_gpsLogHandler = new GpsLogHandler(recentTerm, recentLogNum, obsolateTerm);
	_gpsLogHandler->setLogFile(config.readEntry("logDir") + "/" + "gps.log");
	_gpsLogHandler->setDevFile(config.readEntry("devFile"));

	_buffer = new QImage();
	_mapBuffer = new QPixmap();

	// ̹ޡ
	_updateTimer = new QTimer( this );
	connect(_updateTimer, SIGNAL(timeout()), this, SLOT(slotRePaint()) );

	// GPS¬μư󥿥󥰵ǽͭ
	_autoCentered = true;
}

Zega::~Zega() {
	if (_gpsLogHandler->isAlive())
		stopGps();
	
	config.setGroup("CurrentCoordinate");
	config.writeEntry("lat", _centerNS.toNS());
	config.writeEntry("lon", _centerEW.toEW());
	config.setGroup("Maps");
	config.writeEntry("currentMap", (int)_mapIndex);
}

void Zega::slotChScale(uint mapId) {
	_mapIndex = mapId;
	_mapList->changeMap(mapDirs[mapId]);
	paintMaps(_centerNS, _centerEW);
}

void Zega::slotRePaint() {
	static int beforeX = -50;
	static int beforeY = -50;
	Coordinate coordinate;

	if (_gpsLogHandler->isAlive() == false) {
		QPainter painter(this);
		painter.setPen(QPen(black, 1, SolidLine));
		painter.setBrush(QBrush(black, SolidPattern));
		painter.drawEllipse(10, 40, 10, 10);
		painter.end();
		return;
	}

	// ּ߰
	coordinate = _gpsLogHandler->getNowCoordinate();
	double direction = _gpsLogHandler->getNowDirection();
	int x = (coordinate.getRadEW().toLong() - _centerEW.toLong()) / _pixelEW + this->width()/2;
	int y = (coordinate.getRadNS().toLong() - _centerNS.toLong()) / _pixelNS + this->height()/2;

	// ֤߰ϰϳξϸ֤߰ѹ
	if (_autoCentered) {
		if ( x < 50 || x > this->width() -50 || y < 50 || y > this->height() -50) {
			_centerNS = coordinate.getRadNS();
			_centerEW = coordinate.getRadEW();
			paintMaps(_centerNS, _centerEW);
			x = (coordinate.getRadEW().toLong() - _centerEW.toLong()) / _pixelEW + this->width()/2;
			y = (coordinate.getRadNS().toLong() - _centerNS.toLong()) / _pixelNS + this->height()/2;
		}
	}

	// 
	if (beforeX >= 0 && beforeX < this->width() && beforeY >= 0 && beforeY < this->height()) {
		bitBlt(this, beforeX-10, beforeY-10, _mapBuffer,
				beforeX-10, beforeY-10, 21, 21);
	}
	
	// ֤߰Ͽޤ˥ݥ
	if (x >= 0 && x < this->width() && y >= 0 && y < this->height()) {
		// ϰϵϿ
		beforeX = x;
		beforeY = y;

		// ֤߰ޤǤεפХåե˵Ͽ
		QPainter painter(_mapBuffer);
		painter.setPen(QPen(blue, 1, SolidLine));
		painter.setBrush(QBrush(white, SolidPattern));
		painter.drawRect(x -2, y -2, 5, 5);
		painter.end();

		// 		
		painter.begin(this);
		painter.translate(x, y);
		painter.setPen(QPen(QColor(44, 0, 44, QColor::Rgb), 1, SolidLine));
		painter.setBrush(QBrush(green, SolidPattern));
		QPointArray tempPoints(4);
		tempPoints.setPoint(0, rotate(QPoint(0, 10), direction));
		tempPoints.setPoint(1, rotate(QPoint(-5, -9), direction));
		tempPoints.setPoint(2, 0, 0);
		tempPoints.setPoint(3, rotate(QPoint(5, -9), direction));
		painter.drawPolygon(tempPoints);
		painter.end();
	}
	QPainter painter(this);
    // GPS򥢥ƥ֤ɽ
	painter.setPen(QPen(red, 1, SolidLine));
	painter.setBrush(QBrush(red, SolidPattern));
	painter.drawEllipse(10, 40, 10, 10);
	// ־ɽ
	painter.setPen(QPen(black, 1, SolidLine));
	painter.setBrush(QBrush(white, SolidPattern));
	painter.setFont(QFont("courier", 12));
	painter.drawRect(this->width() -150, 40, 140, 44);
	painter.drawText(this->width() -148, 54, coordinate.getRadNS().toNS());
	painter.drawText(this->width() -148, 68, coordinate.getRadEW().toEW());
	painter.drawText(this->width() -148, 82, "Direction: " + QString::number(direction, 'f', 2));
	painter.end();
}

void Zega::startGps() {
	config.setGroup("GPS");
	if ( !config.hasKey("UpdateInterval") ) {
		config.writeEntry("UpdateInterval", 1000);
	}
	int intInterval = config.readNumEntry("UpdateInterval");

	_gpsLogHandler->start();
	_updateTimer->start(intInterval, FALSE);
}

void Zega::stopGps() {
	_updateTimer->stop();
	_gpsLogHandler->stop();
	cout << "Cardctl end." << endl;
}

void Zega::slotLoad() {
	QTKFileDialog dialog(this, "Load KML File");
	dialog.setFilter(tr("All|*.kml *.mrd"));

	config.setGroup("Data");
	if (!config.hasKey("dir"))
		config.writeEntry("dir", "/mnt/card");

	dialog.setDirectory(config.readEntry("dir"));
	dialog.setDirectorySelectMode(false);
	dialog.setHomeDirectoryCdUpEnable(true);

	if (dialog.exec() == QDialog::Accepted) {
		QFile routeFile;
		bool loadKml = false;
		if (dialog.getFileName().right(4) == ".kml") {
			loadKml = true;
		}
		routeFile.setName(dialog.getFileName());
		if (routeFile.exists()) {
			if (!routeFile.open(IO_ReadOnly)) {
				QMessageBox::warning(this, "Load error",
						"Could not load file.\n" + dialog.getFileName());
				return;
			}
		} else {
			QMessageBox::warning(this, "Load error", "File not found.\n"
					+ dialog.getFileName());
			return;
		}
		QTextStream qtsIn(&routeFile);

		if (loadKml) {
			while (!qtsIn.eof()) {
				QString str = qtsIn.readLine();
				if (str.contains("<coordinates>", false) > 0)
					break;
			}
		}

		_route.clear();
		Radian radNS, radEW;
		while (!qtsIn.eof()) {
			QString str = qtsIn.readLine();
			QStringList qslTemp = QStringList::split(",", str);
			int intHight = 0;
			if (loadKml && qslTemp.count() >= 3) {
				intHight = qslTemp[2].toInt();
			}
			if (qslTemp.count() >= 2) {
				radEW.parse2("E" + qslTemp[0]);
				radNS.parse2("N" + qslTemp[1]);
				Coordinate coordinate(radEW, radNS, intHight);
				if (loadKml)
					coordinate.convWGS84ToTokyo();
				_route.push_back(coordinate);
			}
		}
		routeFile.close();

		// ̺
		paintMaps(_centerNS, _centerEW);
	}
}

void Zega::resizeEvent(QResizeEvent *event) {
	if (_mapBuffer->width() != this->width() || _mapBuffer->height()
			!= this->height()) {
		_mapBuffer->resize(this->width(), this->height());
		paintMaps(_centerNS, _centerEW);
	}
}

void Zega::mousePressEvent(QMouseEvent *event) {
#ifdef _DEBUG_PRINT
	cout << "mousePressEvent " << event->pos().x() << ", " << event->pos().y() << endl;
#endif

	Radian radNS, radEW;

	radNS.setData(_centerNS.toLong() + _pixelNS * (event->pos().y() - _mapBuffer->height()/2));
	radEW.setData(_centerEW.toLong() + _pixelEW * (event->pos().x() - _mapBuffer->width()/2));

	paintMaps(radNS, radEW);
}

void Zega::keyPressEvent(QKeyEvent *event) {
	Radian radNS, radEW;
	int intWait = 2;

	radNS = _centerNS;
	radEW = _centerEW;

	if (event->state() & ShiftButton) {
		intWait = 1;
	}

	if (event->key() == Key_Up) {
		radNS.setData(_centerNS.toLong() - _pixelNS * (_mapBuffer->height()
				/intWait));
	} else if (event->key() == Key_Down) {
		radNS.setData(_centerNS.toLong() + _pixelNS * (_mapBuffer->height()
				/intWait));
	} else if (event->key() == Key_Left) {
		radEW.setData(_centerEW.toLong() - _pixelEW * (_mapBuffer->width()
				/intWait));
	} else if (event->key() == Key_Right) {
		radEW.setData(_centerEW.toLong() + _pixelEW * (_mapBuffer->width()
				/intWait));
	} else if (event->key() == Key_Space) {
		if (_gpsLogHandler->isAlive()) {
			Coordinate coordinate = _gpsLogHandler->getNowCoordinate();
			radNS = coordinate.getRadNS();
			radEW = coordinate.getRadEW();
		} else {
			return;
		}
	} else {
		switch (event->key()) {
		case Key_1:
			if (mapDirs.count() > 0)
				slotChScale(0);
			break;
		case Key_2:
			if (mapDirs.count() > 1)
				slotChScale(1);
			break;
		case Key_3:
			if (mapDirs.count() > 2)
				slotChScale(2);
			break;
		case Key_4:
			if (mapDirs.count() > 3)
				slotChScale(3);
			break;
		case Key_5:
			if (mapDirs.count() > 4)
				slotChScale(4);
			break;
		case Key_6:
			if (mapDirs.count() > 5)
				slotChScale(5);
			break;
		case Key_G:
			if (_gpsLogHandler->isStopped())
				startGps();
			else
				stopGps();
			break;
		case Key_C:
			if (_autoCentered)
				_autoCentered = false;
			else
				_autoCentered = true;
			break;
		case Key_L:
			slotLoad();
			break;
		case Key_D:
			_gpsLogHandler->changeDebugPrint();
			break;
		default:
			break;
		}
		return;
	}

#ifdef _DEBUG_PRINT
	QString tempStr;
	tempStr = radEW.toEW();
	cout << "point EW: " << tempStr << endl;
	tempStr = radNS.toNS();
	cout << "point NS: " << tempStr << endl;
#endif

	paintMaps(radNS, radEW);
}

void Zega::paintEvent(QPaintEvent*) {
	if (_mapBuffer->width() == 0) {
		_mapBuffer->resize(this->width(), this->height());
		paintMaps(_centerNS, _centerEW);
	} else {
		bitBlt(this, 0, 0, _mapBuffer);
	}
}

// ץɽ
void Zega::about() {
	QMessageBox::about(
			this,
			"About this program",
			"Zega Version 0.1\nThis is a GPS assistance program for Zaurus.\nCopyright HB1 2008.05.01");
}

void Zega::paintMaps(Radian radNS, Radian radEW) {
	Radian radMapNS, radMapEW;
	Radian radMapWidth, radMapHeight;
	Radian radMinNS, radMinEW;
	QString mapFileName;
	QString tempStr;

	// 濴ηٰ٤Ͽ
	_centerNS = radNS;
	_centerEW = radEW;

	// Ͽ޽񤭹ߥХåեꥢ
	_mapBuffer->fill();

	mapFileName = _mapList->searchMapFileName(radNS.toLong(), radEW.toLong());
	if (getMapInfo(mapFileName, &radMapEW, &radMapNS, &radMapWidth,
			&radMapHeight)) {
		radMinEW.setData(radMapEW.toLong() - radMapWidth.toLong()/2);
		radMinNS.setData(radMapNS.toLong() - radMapHeight.toLong()/2);

#ifdef _DEBUG_PRINT
		tempStr = radEW.toEW();
		cout << "point EW: " << tempStr << endl;
		tempStr = radNS.toNS();
		cout << "point NS: " << tempStr << endl;
		tempStr = radMinEW.toEW();
		cout << "lower EW: " << tempStr << endl;
		tempStr = radMinNS.toNS();
		cout << "lower NS: " << tempStr << endl;
#endif
		if (!_buffer->load(mapFileName)) {
			QMessageBox::warning(this, "Load error", "Could not load file.");
			return;
		}

		int x = (radEW.toLong() - radMinEW.toLong())*(_buffer->width())
				/radMapWidth.toLong();
		int y = (radNS.toLong() - radMinNS.toLong())*(_buffer->height())
				/radMapHeight.toLong();
		int isx = x - _mapBuffer->width()/2;
		_bsl = 0;
		if (isx < 0) {
			_bsl = _mapBuffer->width()/2 - x;
			isx = 0;
		}
		int isy = y - _mapBuffer->height()/2;
		_bsu = 0;
		if (isy < 0) {
			_bsu = _mapBuffer->height()/2 - y;
			isy = 0;
		}

		// Ͽ޲ԡ
		bitBlt(_mapBuffer, _bsl, _bsu, _buffer, isx, isy, _buffer->width()
				- isx, _buffer->height() - isy);

		// ;׻
		_bsr = _bsl + _buffer->width() - isx;
		_bsd = _bsu + _buffer->height() - isy;

		// 1 pixel ٰ٤Ͽ
		_pixelNS = radMapHeight.toLong() / _buffer->height();
		_pixelEW = radMapWidth.toLong() / _buffer->width();
	} else {
		_bsl = _mapBuffer->width();
		_bsu = _mapBuffer->height();
		_bsr = 0;
		_bsd = 0;
	}

	copyMap( 1, 0, radEW, radNS);
	copyMap(-1, 0, radEW, radNS);
	copyMap( 0, 1, radEW, radNS);
	copyMap( 0, -1, radEW, radNS);
	copyMap( 1, 1, radEW, radNS);
	copyMap( 1, -1, radEW, radNS);
	copyMap(-1, 1, radEW, radNS);
	copyMap(-1, -1, radEW, radNS);

	// ּ߰
	Coordinate coordinate;
	int x, y, x2, y2;
	long cEW = _centerEW.toLong();
	long cNS = _centerNS.toLong();
	int cx = this->width()/2;
	int cy = this->height()/2;
	long lEW = _centerEW.toLong() - _pixelEW * cx;
	long uEW = _centerEW.toLong() + _pixelEW * cx;
	long lNS = _centerNS.toLong() - _pixelNS * cy;
	long uNS = _centerNS.toLong() + _pixelNS * cy;
	QPainter painter(_mapBuffer);

	// 롼Ȥ
	bool writeRoute = false;
	painter.setPen(QPen(red, 3, SolidLine));
	vector<Coordinate>::iterator it = _route.begin();
	if (it != _route.end()) {
		x = ((*it).getRadEW().toLong() - cEW) / _pixelEW + cx;
		y = ((*it).getRadNS().toLong() - cNS) / _pixelNS + cy;
	}
	while (it != _route.end()) {
		x2 = ((*it).getRadEW().toLong() - cEW) / _pixelEW + cx;
		y2 = ((*it).getRadNS().toLong() - cNS) / _pixelNS + cy;
		if (writeRoute || ((*it).getRadEW().toLong() >= lEW && (*it).getRadEW().toLong() <= uEW && (*it).getRadNS().toLong() >= lNS && (*it).getRadNS().toLong() <= uNS)) {
			painter.drawLine(x, y, x2, y2);
			writeRoute = true;
		}
		x = x2;
		y = y2;
		it++;
	}

	// ľʳΥݥ
	painter.setPen(QPen(blue, 1, SolidLine));
	painter.setBrush(QBrush(white, SolidPattern));
	_gpsLogHandler->setObsolateLogFirst();
	while (!_gpsLogHandler->endObsolateLog()) {
		coordinate = _gpsLogHandler->getNextObsolateLog();

		// ֤Ͽޤ˥ݥ
		if (coordinate.getRadEW().toLong() >= lEW && coordinate.getRadEW().toLong() <= uEW && coordinate.getRadNS().toLong() >= lNS && coordinate.getRadNS().toLong() <= uNS) {
			x = (coordinate.getRadEW().toLong() - cEW) / _pixelEW + cx;
			y = (coordinate.getRadNS().toLong() - cNS) / _pixelNS + cy;
			painter.drawRect(x -2, y -2, 5, 5);
		}
	}

	// ľΥݥ
	_gpsLogHandler->setRecentLogFirst();
	while (!_gpsLogHandler->endRecentLog()) {
		coordinate = _gpsLogHandler->getNextRecentLog();

		// ֤Ͽޤ˥ݥ
		if (coordinate.getRadEW().toLong() >= lEW && coordinate.getRadEW().toLong() <= uEW && coordinate.getRadNS().toLong() >= lNS && coordinate.getRadNS().toLong() <= uNS) {
			x = (coordinate.getRadEW().toLong() - cEW) / _pixelEW + cx;
			y = (coordinate.getRadNS().toLong() - cNS) / _pixelNS + cy;
			painter.drawRect(x -2, y -2, 5, 5);
		}
	}

	painter.setPen(QPen(black, 1, SolidLine));
	painter.drawLine(_mapBuffer->width()/2 - 10, _mapBuffer->height()/2,
			_mapBuffer->width()/2 + 10, _mapBuffer->height()/2);
	painter.drawLine(_mapBuffer->width()/2, _mapBuffer->height()/2 - 10,
			_mapBuffer->width()/2, _mapBuffer->height()/2 + 10);
	bitBlt(this, 0, 0, _mapBuffer);
}

// Ͽ޾
bool Zega::getMapInfo(QString mapFileName, Radian *radMapEW, Radian *radMapNS,
		Radian *radMapWidth, Radian *radMapHeight) {
	if (mapFileName.isEmpty())
		return false;

	int slashPoint = mapFileName.findRev('/');
	int dotPoint = mapFileName.findRev('.');
	QString tempStr = mapFileName.mid(slashPoint +1, dotPoint - slashPoint - 1);

	QStringList mapInfo = QStringList::split(",", tempStr);
	if (mapInfo.count() < 4)
		return false;

	radMapNS->parse(mapInfo[0]);
	radMapEW->parse(mapInfo[1]);
	if (mapInfo[2].find('.') == -1)
		mapInfo[2] += ".0";
	radMapHeight->parse("S0.0." + mapInfo[2]);
	if (mapInfo[3].find('.') == -1)
		mapInfo[3] += ".0";
	radMapWidth->parse("E0.0." + mapInfo[3]);

	if (radMapWidth->toLong() == 0 || radMapHeight->toLong() == 0)
		return false;

	return true;
}

// Ͽɹ
bool Zega::copyMap(int x, int y, Radian radEW, Radian radNS) {
	Radian radMapEW, radMapNS;
	Radian radMinEW, radMinNS;
	Radian radMapWidth, radMapHeight;
	QString mapFileName;

	if (x < 0 && _bsl <= 0)
		return false;
	if (x > 0 && _bsr >= _mapBuffer->width())
		return false;
	if (y < 0 && _bsu <= 0)
		return false;
	if (y > 0 && _bsd >= _mapBuffer->height())
		return false;

	mapFileName = _mapList->pointMapFileName(y, x);

	if (getMapInfo(mapFileName, &radMapEW, &radMapNS, &radMapWidth,
			&radMapHeight)) {
		radMinEW.setData(radMapEW.toLong() - radMapWidth.toLong()/2);
		radMinNS.setData(radMapNS.toLong() - radMapHeight.toLong()/2);

		if (!_buffer->load(mapFileName)) {
			QMessageBox::warning(this, "Load error", "Could not load file.");
			return false;
		}

		int w = (radEW.toLong() - radMinEW.toLong())*(_buffer->width())
				/radMapWidth.toLong();
		int h = (radNS.toLong() - radMinNS.toLong())*(_buffer->height())
				/radMapHeight.toLong();
		int isx = w - _mapBuffer->width()/2;
		int bsx = 0;
		if (isx < 0) {
			bsx = _mapBuffer->width()/2 - w;
			isx = 0;
		}
		int isy = h - _mapBuffer->height()/2;
		int bsy = 0;
		if (isy < 0) {
			bsy = _mapBuffer->height()/2 - h;
			isy = 0;
		}

		int copy_width = _buffer->width() - isx;
		if (x > 0 && _bsr > bsx) {
			isx += _bsr - bsx;
			copy_width -= _bsr - bsx;
			bsx = _bsr;
		} else if (x < 0 && copy_width > _bsl - bsx)
{
			copy_width = _bsl - bsx;
		}

    				int copy_height = _buffer->height() - isy;
		if (y > 0 && _bsd > bsy) {
			isy += _bsd - bsy;
			copy_height -= _bsd - bsy;
			bsy = _bsd;
		} else if (y < 0 && copy_height > _bsu - bsy)
{
			copy_height = _bsu - bsy;
		}

		// Ͽ޲ԡ
					bitBlt(_mapBuffer, bsx, bsy, _buffer, isx, isy, copy_width,
					copy_height);
	}
	return true;
}

